import os
import json
import logging
import re
from uuid import uuid1

from croniter import croniter
from datetime import datetime, date, timedelta
from pprint import pprint

from p2.agent.client import Client as AgentClient
from p2.agent.cronjob import Cronjob
from p2.store.client import Client as StoreClient
from p2.scheduler.Config import Config
from p2.scheduler.Calendar import Calendar
from p2.scheduler.Test import P2test
from p2.scheduler.k8.resources import Resources
from p2.logging import Logger


class Scheduler:

    def __init__(self, logger=None, context:str=None):

        try:
            self.log = logger.getLogger() if logger else logging.getLogger(__name__)

            store_url = os.environ.get('P2_STORE_URL')
            agent_url = os.environ.get('P2_AGENT_URL')
            self.log.debug(f"Store URL {store_url}")
            self.log.debug(f"Agent URL {agent_url}")

            self.p2_store_client = StoreClient(store_url)
            if self.p2_store_client is None:
                raise Exception("Error getting a p2.store client object")

            self.p2_agent_client = AgentClient(agent_url)
            if self.p2_agent_client is None:
                raise Exception("Error getting a p2.agent client object")

            self.config = Config()
            self.to_be_scheduled = list()

            if context:
                self.context = context

        except Exception as err:
            self.log.error(err)
            return None

    def schedule_campaign(self, campaign_id: int):
        """
        - retrieves the campaign with id
        - get targets
        - get experiments
        - see if experiments have to be scheduled...
        """
        self.campaign_id = campaign_id
        campaign = self.p2_store_client.get_campaign_of_id(campaign_id)
        if campaign is None:
            raise Exception(f"No campaign with id {id} found")

        if campaign['resources'] is None:
            raise Exception(f"Error, no resources set for campaign {id}")

        if campaign['experiments'] is None:
            raise Exception(f"Error, no experiments set for campaign {id}")

        context = campaign['resources'].get('context', 'minikube')
        targets = self.p2_store_client.get_k8s_resources(campaign['resources'])
        experiments = self.p2_store_client.get_experiments(campaign['experiments'])
        self.cal = Calendar(context)

        # Find out if new resources have appeared since last scheduling
        resources_obj = Resources(context=context)
        resources_obj.update()

        for target in targets:
            for experiment in experiments:
                if  experiment['type'] == 'experiment' and \
                    'profile' in target and \
                    'target' in experiment['profile'] and \
                    target['profile']['name'] == experiment['profile']['target']['name']:
                    self._add_to_schedule(target=target, experiment=experiment)
                elif experiment['type'] == 'template':
                    # if it's a template then any target is valid...
                    self._add_to_schedule(
                        campaign=campaign,
                        target=target,
                        experiment=experiment
                    )

        self._schedule_all()
        return targets

    def _add_to_schedule(self, target: dict, experiment: dict, campaign: dict=None):
        self.log.info("See if experiment {} on target {} has to be scheduled...".format(
            experiment['filename'], target['name']))

        target_priority=self._get_target_priority( target )

        # * get max(next_run_time), status from session
        # where filepath=experiment and profile[‘target’][‘name’]=resource 
        self.log.info(f"Getting all sessions executing experiment {experiment['filename']}...")
        sessions = self.p2_store_client.get_sessions(
            filepath=experiment['filename'],
            context=target['context']
            )

        valid_sessions = [ s for s in sessions if _valid_session(s, target) ]
        if len(valid_sessions):
            sessions_sorted = sorted(valid_sessions, key=lambda session: session['next_run_time'])
            last_session = sessions_sorted[-1]
            min_interval = self.min_interval(target, experiment, last_session['status'])
            expired_by = timedelta(days=min_interval+1)
            if 'end_time' in last_session and last_session['end_time'] != None:
                expired_by = date.today() - datetime.fromisoformat(last_session['end_time']).date()

            if expired_by.days > min_interval:
                self.to_be_scheduled.append( P2test(
                        campaign = campaign,
                        target = target,
                        experiment = experiment,
                        expired = expired_by.days - min_interval))
                return True
        else:
            self.to_be_scheduled.append( P2test(
                campaign = campaign,
                target = target,
                experiment = experiment,
                expired = self.config.target_priority_period[target_priority-1]
                ))
            return True
        return False

    def schedule(self):
        """schedules all experiments that need to be scheduled...
        Note: this has been superseeded by schedule_campaign..."""

        # gets list of targets that match input params
        # TODO : get only resources in active status...
        targets = self.get_resources()
        for target in targets:
            self.log.info("Found target {}".format(target['name']))

            # * get all experiments on this target or target kind:
            experiments = self.get_experiments(target=target)

            # * for each experiment:
            for experiment in experiments:
                self._add_to_schedule(target=target, experiment=experiment)

        return self._schedule_all()

    def _schedule_all(self):
        """schedule tests sorted by expiretion time"""
        not_scheduled=list()
        to_be_scheduled_sorted = sorted(self.to_be_scheduled, key=lambda item: item.expired)
        for test in to_be_scheduled_sorted[::-1]:
            for day in self.cal.calendar:
                if len(self.cal.calendar[day]['sessions']) < self.config.max_tests_per_day:
                    self._schedule_this(test, day)
                    break;
            else:
                self.log.warning("No days left to schedule experiment {} on target {}".format(
                    test.experiment['filename'], test.target['name']))
                not_scheduled.append(test)
        return not_scheduled

    def _schedule_this(self, test: P2test, day: date):
        """finds the best slot in the day for this test"""
        """ Gets the max(next_run_time) of the day
            adds that test expected duration and round up to next hour
            check hour is in test shift"""
        """ in case of a template:
            - template profile should hold the template static parameters
            in case these are set in the test campaign, they override the ones
            set in the template. This is handled by the agent (cron) job.
            - session profile should hold the dynamic params
            - Current dynamic params supported:
                - target: target profile
            """
        try:
            last_time = self._get_last_time(day)
            duration = test.experiment['duration'] if 'duration' in test.experiment else 30
            test_time = last_time + timedelta(minutes=duration+30)
            expected_end_time = test_time + timedelta(minutes=test.experiment['duration'])
            if (expected_end_time.hour + 1) in self.config.hours_slot:
                mins = "%02d" % test_time.minute
                hour = "%02d" % test_time.hour
                dom  = "%02d" % test_time.day
                month= "%02d" % test_time.month
                schedule = f"{mins} {hour} {dom} {month} *"
                cron = croniter(schedule, datetime.now())
                next_run_time = cron.get_next(datetime)
                exp_id = test.target['name'] + '-' + str(uuid1())[:8]
                session = {
                        'filepath': test.experiment['filename'],
                        'campaign_id': self.campaign_id,
                        'context': test.target['context'],
                        'schedule': schedule,
                        'next_run_time': str(next_run_time),
                        'mins': mins,
                        'hour': hour,
                        'dom': dom,
                        'month': month,
                        'dow': '*',
                        'year': test_time.year,
                        'profile':{'target':test.target},
                        'status': 'loaded'
                }
                exp_id = self.p2_store_client.create_session(exp_id, session)
                if exp_id != None:
                    cronjob = Cronjob(
                            id = exp_id,
                            crontab = schedule,
                            filename = test.experiment['filename'],
                            driver = test.experiment['type'],
                            campaign_id = self.campaign_id
                            )
                    ret = self.p2_agent_client.create_cronjob(test.target['context'], cronjob)
                    if ret != exp_id:
                        raise Exception("Couldn't add cronjob")

                    self.p2_store_client.update_session(exp_id, {'status':'scheduled'})

                    if test in self.to_be_scheduled:
                        self.to_be_scheduled.remove(test)

                    self.cal.calendar[test_time.date()]['sessions'].append(session)
                    self.log.info("Session {} scheduled".format(exp_id))
                else:
                    self.log.error(
                        "Error scheduling session for experiment {}".format(session['filepath']))
            else:
                self.log.warning("Couldn't schedule test {}, not enough time left in the day")
            return exp_id
        except Exception as e:
            self.log.error(f"Couldn't schedule test, exception: {e}")
            return None

    def _get_last_time(self, day):
        if day not in self.cal.calendar:
            raise Exception(f"Day {day} not in calendar!")
        last_time = datetime(day.year, day.month, day.day, self.config.hours_slot[0], 0)
        for session in self.cal.calendar[day]['sessions']:
            next_run_time = datetime.fromisoformat(session['next_run_time'])
            if next_run_time > last_time:
                last_time = next_run_time
        return last_time

    def get_experiments(self, target: dict):
        all = self.p2_store_client.execute_command(
                cmd="select * from experiments",
                attrs=[]
                )
        valid = list()
        for experiment in all:
            if (('profile' in experiment and experiment['profile'] is not None) and
                ('target' in experiment['profile']) and
                (experiment['profile']['target']['kind'] == target['kind']) and
                (experiment['profile']['target']['name'] == target['name'])):
                valid.append(experiment)
        return valid

    def get_resources(self):
        resources = self.p2_store_client.get_k8s_resources(
                context=self.context,
                status='active',
                kind=self.kind)
        if self.target_pattern is None: return resources

        targets = list()
        #filter only resources that match pattern
        pattern = re.compile(self.target_pattern)
        for resource in resources:
            if pattern.match(resource['name']):
                targets.append(resource)

        # TODO filter if target_profile is set.
        return targets

    def min_interval(self, target, experiment, status):
        """compoutes min interval (in days ) between tests
            of same type on same target """

        exp_profile = experiment.get('profile',{'weight':1.0})
        target_priority = self._get_target_priority(target)
        weight = exp_profile.get('weight', 1.0)
        if status == 'failed':
            if weight > 0.9: weight = weight - 0.5 
        return int(self.config.target_priority_period[target_priority-1] * weight)

    def _get_target_priority(self, target):
        # target_priority=len(self.config.target_priority_period) or 1
        target_priority = 1
        if target['context'] in self.config.targets_priority:
            target_priority=self.config.targets_priority[target['context']].get(target['name'],
                    len(self.config.target_priority_period))
        return target_priority

def _valid_session(session: dict, target: dict):
    if (    'profile' not in session
            or session['profile'] == None
            or 'target' not in session['profile']
            or 'name' not in session['profile']['target']
            or session['profile']['target']['name'] != target['name']):
        return False
    else:
        return True
