"""A p2.scheduler class to deal with k8 resources.

Usage:
    from p2.scheduler.k8.resources import Resources

    obj = Resources(context=context)

    # Updates catalogue or k8 resources on p2.store
    obj.update()
"""
import os
import json
import logging
from datetime import datetime

from p2.driver.k8s import k8s
from p2.store.client import Client
from p2.logging import Logger


class Resources:

    def __init__(self, context:str, logger:Logger=None):
        self.context = context
        self.k8s_api = k8s.Api(context=context)
        self.log = logger.getLogger() if logger else logging.getLogger(__name__)
        self.p2_store_url = os.environ.get('P2_STORE_URL')
        if self.p2_store_url is None:
            raise Exception("P2_STORE_URL not set")
        self.client = Client(self.p2_store_url, timeout=10)

    def update(self):
        """
        Updates the list of k8 resources stored in p2.store for the context under test
        """

        namespaces = self.k8s_api.namespaces()

        for namespace in namespaces:
            record = {
                'context': self.context,
                'kind': 'namespace',
                'name': namespace['name'],
                'profile': json.dumps(namespace),
                'status': 'active',
                'last_updated': str(datetime.now())
            }
            ret = self.client.update_on_conflict(
                table='k8s_resources',
                conflicting_cols=['context', 'kind', 'name'],
                record=record
                )

            if ret is not None:
                self.log.info(f"Resource {record['name']} updated")
            else:
                self.log.error(f"Updating resource {record['name']}")
                raise Exception(f"Error updating resource {record['name']}")

            namespace_rec = self.client.select(
                table='k8s_resources',
                filter= {
                    'context': self.context,
                    'kind': 'namespace',
                    'name': record['name'],
                },
                limit=1
            )
            if len(namespace_rec) == 0:
                self.log.error(f"Something went wrong with resource {record['name']}")
                raise Exception(f"Something went wrong with resource {record['name']}")

            deployments = self.k8s_api.deployments_for_namespace(namespace['name'])
            for name in deployments.keys():
                healthy = self.k8s_api.is_deployment_healthy(deployments[name])
                record = {
                    "parent_id": namespace_rec[0]['id'],
                    "context": self.context,
                    "kind": "deployment",
                    "name": name,
                    "profile": json.dumps(deployments[name]),
                    "status": 'healthy' if healthy else 'not healthy',
                    "last_updated": str(datetime.now())
                }
                ret = self.client.update_on_conflict(
                    table='k8s_resources',
                    conflicting_cols=['context', 'kind', 'name'],
                    record=record
                    )

        # ...and now check if there are resources in p2.store that are not any more
        # there and mark them as deleted...
        resources = self.client.select(
            table = 'k8s_resources',
            filter = {
                'context': self.context,
                'kind': 'namespace'
            },
            limit = 10000)

        if (len(resources) > 0):
            for resource in resources:
                if resource['name'] not in [ns['name'] for ns in namespaces]:
                    self.log.info(f"""Namespace {resource['name']} not found on context
                                  {self.context}, marking it as deleted...""")
                    resource['status'] = 'deleted'
                    for k in ['id', 'parent_id', 'profile']:
                        del(resource[k])
                    record['last_updated'] = str(datetime.now())
                    self.client.update_on_conflict(
                        table = 'k8s_resources',
                        record = resource,
                        conflicting_cols = ['context', 'kind', 'name']
                    )

        return 1
